#!/usr/bin/env node
"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const child_process_1 = __importDefault(require("child_process"));
const fs_1 = __importDefault(require("fs"));
const promises_1 = __importDefault(require("fs/promises"));
const chalk_1 = __importDefault(require("chalk"));
const boxen_1 = __importDefault(require("boxen"));
const yargs_1 = __importDefault(require("yargs/yargs"));
const discord_js_1 = __importDefault(require("discord.js"));
const string_similarity_1 = __importDefault(require("string-similarity"));
const figlet_1 = __importDefault(require("figlet"));
const loading_cli_1 = __importDefault(require("loading-cli"));
const path_1 = require("path");
const helpers = require("yargs/helpers");
const events = require("../events.json");
const exec = (cmd, options) => {
    return new Promise((res, rej) => {
        child_process_1.default.exec(cmd, options, (err) => {
            if (err)
                rej(err);
            else
                res(null);
        });
    });
};
const root = (...segments) => (0, path_1.join)(process.cwd(), ...segments);
async function readJSON(path) {
    return JSON.parse(await promises_1.default.readFile(path, "utf8"));
}
async function writeJSON(path, json) {
    await promises_1.default.writeFile(path, JSON.stringify(json, null, 2), "utf8");
}
async function loader(start, callback, end) {
    const time = Date.now();
    const load = (0, loading_cli_1.default)({
        text: start,
        interval: 150,
        color: "white",
        frames: ["◐", "◓", "◑", "◒"],
    }).start();
    await callback();
    load.succeed(`${end} ${chalk_1.default.grey(`${Date.now() - time}ms`)}`);
}
async function setupDatabase(projectPath, database) {
    const conf = await readJSON((0, path_1.join)(projectPath, "package.json"));
    // delete all other database dependencies.
    for (const dbname of ["sqlite3", "mysql2", "pg"]) {
        if (dbname !== database.database)
            conf.dependencies[dbname] = undefined;
        else
            conf.dependencies[dbname] = "latest";
    }
    await writeJSON((0, path_1.join)(projectPath, "package.json"), conf);
    const template = await promises_1.default.readFile((0, path_1.join)(__dirname, "..", "templates", database.database), "utf8");
    await promises_1.default.writeFile((0, path_1.join)(projectPath, "src", "app", "database.ts"), template, "utf8");
    if (database.host)
        await injectEnvLine("DB_HOST", database.host, projectPath);
    if (database.port)
        await injectEnvLine("DB_PORT", database.port, projectPath);
    if (database.user)
        await injectEnvLine("DB_USER", database.user, projectPath);
    if (database.password)
        await injectEnvLine("DB_PASSWORD", database.password, projectPath);
    if (database.dbname)
        await injectEnvLine("DB_DATABASE", database.dbname, projectPath);
}
async function injectEnvLine(name, value, projectPath = process.cwd()) {
    const env = await promises_1.default.readFile((0, path_1.join)(projectPath, ".env"), "utf8");
    const lines = env.split("\n");
    const index = lines.findIndex((line) => line.split("=")[0] === name);
    if (index > -1)
        lines.splice(index, 1);
    lines.push(`${name}="${value}"`);
    await promises_1.default.writeFile((0, path_1.join)(projectPath, ".env"), lines.join("\n"), "utf8");
}
async function isValidRoot() {
    let conf = null;
    try {
        conf = require(root("package.json"));
    }
    catch (e) { }
    if (!conf ||
        !conf.hasOwnProperty("devDependencies") ||
        !conf.devDependencies.hasOwnProperty("make-bot.ts")) {
        console.error(chalk_1.default.red('you should only use this command at the root of a "bot.ts" project'));
        return false;
    }
    return true;
}
(0, yargs_1.default)(helpers.hideBin(process.argv))
    .scriptName("make")
    .usage("$0 <cmd> [args] [--options]")
    .command("bot [name] [path] [--options]", "create typescript bot", (yargs) => yargs
    .positional("name", {
    default: "bot.ts",
    describe: "bot name",
})
    .positional("path", {
    default: ".",
    describe: "bot path",
})
    .option("code-style", {
    alias: ["style", "code"],
    default: "options",
    choices: ["options", "chain"],
    describe: "style of code",
})
    .option("prefix", {
    alias: "p",
    default: ".",
    describe: "bot prefix",
})
    .option("locale", {
    alias: "l",
    default: "en",
    describe: "locale timezone",
})
    .option("database", {
    alias: "d",
    default: "sqlite3",
    choices: ["sqlite3", "mysql2", "pg"],
    describe: "used database",
})
    .option("token", {
    alias: "t",
    type: "string",
    describe: "bot token",
})
    .option("secret", {
    type: "string",
    describe: "bot secret",
})
    .option("owner", {
    alias: "o",
    type: "string",
    describe: "your Discord id",
})
    .option("host", {
    alias: "h",
    default: "localhost",
    describe: "database host",
})
    .option("port", {
    describe: "database port",
    type: "string",
})
    .option("user", {
    alias: "u",
    type: "string",
    describe: "database user",
})
    .option("password", {
    alias: "pw",
    type: "string",
    describe: "database password",
})
    .option("dbname", {
    alias: "db",
    type: "string",
    describe: "database name",
}), async (args) => {
    const borderNone = {
        top: " ",
        left: " ",
        right: " ",
        bottom: " ",
        topLeft: " ",
        topRight: " ",
        bottomLeft: " ",
        bottomRight: " ",
        horizontal: " ",
        vertical: " ",
    };
    const warns = [];
    console.log((0, boxen_1.default)(chalk_1.default.blueBright(await new Promise((resolve) => (0, figlet_1.default)("bot.ts", (err, value) => {
        if (err)
            resolve("");
        else
            resolve(value);
    }))), {
        float: "center",
        borderStyle: borderNone,
    }));
    console.time("duration");
    const codeStyle = args["code-style"];
    await loader("downloading", () => exec([
        "git clone",
        "--depth=1",
        "--single-branch",
        "--branch=" + (codeStyle === "options" ? "master" : "design"),
        "https://github.com/bot-ts/framework.git",
        `"${root(args.path, args.name)}"`,
    ].join(" ")), "downloaded");
    const project = (...segments) => root(args.path, args.name, ...segments);
    await loader("initializing", async () => {
        const conf = await readJSON(project("package.json"));
        await writeJSON(project("package.json"), { ...conf, name: args.name });
        await promises_1.default.writeFile(project(".env"), "", "utf8");
        const intents = "GUILDS,GUILD_MEMBERS,GUILD_BANS,GUILD_EMOJIS_AND_STICKERS,GUILD_INTEGRATIONS,GUILD_WEBHOOKS,GUILD_INVITES,GUILD_VOICE_STATES,GUILD_PRESENCES,GUILD_MESSAGES,GUILD_MESSAGE_REACTIONS,GUILD_MESSAGE_TYPING,DIRECT_MESSAGES,DIRECT_MESSAGE_REACTIONS,DIRECT_MESSAGE_TYPING";
        await injectEnvLine("BOT_INTENTS", intents, project());
        await injectEnvLine("BOT_PREFIX", args.prefix, project());
        await injectEnvLine("BOT_LOCALE", args.locale, project());
        const client = new discord_js_1.default.Client({ intents: [] });
        if (args.token) {
            try {
                await client.login(args.token);
            }
            catch (error) {
                return console.error(chalk_1.default.red(`Invalid token given.`));
            }
            await injectEnvLine("BOT_TOKEN", args.token, project());
        }
        if (!client.isReady())
            return console.error(chalk_1.default.red("Discord Client connection error"));
        if (args.token && !args.owner) {
            const app = await client.application.fetch();
            const ownerID = app.owner instanceof discord_js_1.default.User
                ? app.owner.id
                : app.owner?.id ?? "none";
            if (ownerID === "none")
                warns.push("failure to detect bot owner.");
            await injectEnvLine("BOT_OWNER", ownerID, project());
            client.destroy();
        }
        else if (args.owner) {
            await injectEnvLine("BOT_OWNER", args.owner, project());
        }
        if (args.secret)
            await injectEnvLine("BOT_SECRET", args.secret, project());
        await setupDatabase(project(), args);
        await promises_1.default.writeFile(project("readme.md"), `# ${args.name[0].toUpperCase() + args.name.slice(1)} - powered by [bot.ts](https://github.com/CamilleAbella/bot.ts)`);
        await exec("git fetch --unshallow origin", { cwd: project() });
        await exec("git remote remove origin", { cwd: project() });
    }, "initialized");
    await loader("installing", () => exec("npm i", { cwd: project() }), "installed");
    console.log(chalk_1.default.green(`\n${args.name} bot has been created.`));
    console.log(chalk_1.default.cyanBright(`=> ${project()}`));
    console.timeEnd("duration");
    const $ = chalk_1.default.grey("$");
    console.log((0, boxen_1.default)([
        "",
        chalk_1.default.grey("# to quickly create a new file"),
        "  " + $ + " make command [name]",
        "  " + $ + " make listener [ClientEvent]",
        "  " + $ + " make namespace [name]",
        "  " + $ + " make table [name]",
        "",
        chalk_1.default.grey("# to watch typescript and reload " + args.name),
        "  " + $ + " npm run watch",
        "",
        chalk_1.default.grey("# to build typescript and start " + args.name),
        "  " + $ + " npm run start",
        "",
        chalk_1.default.grey("# to simply start " + args.name),
        "  " + $ + " node .",
        "",
        chalk_1.default.grey("# format your files with prettier"),
        "  " + $ + " npm run format",
        "",
    ].join("\n"), {
        float: "center",
        borderStyle: borderNone,
    }));
    warns.forEach(console.warn);
    console.log((0, boxen_1.default)(`Check the validity of the ${chalk_1.default.blueBright(".env")} information. ${chalk_1.default.green("Enjoy!")}`, {
        borderStyle: "round",
        borderColor: "yellow",
        float: "center",
        padding: 1,
    }));
    process.exit(0);
})
    // @ts-ignore
    .command(...makeFile("command", "name"))
    // @ts-ignore
    .command(...makeFile("listener", "event"))
    .command("database <database>", "setup database", (yargs) => {
    yargs
        .positional("database", {
        describe: "used database",
        type: "string",
        choices: ["sqlite3", "mysql2", "pg"],
    })
        .option("host", {
        alias: "h",
        default: "localhost",
        describe: "database host",
    })
        .option("port", {
        describe: "database port",
        type: "string",
    })
        .option("user", {
        alias: "u",
        type: "string",
        describe: "database user",
    })
        .option("password", {
        alias: "pw",
        type: "string",
        describe: "database password",
    })
        .option("dbname", {
        alias: "db",
        type: "string",
        describe: "database name",
    });
}, async (args) => {
    console.time("duration");
    // @ts-ignore
    await setupDatabase(root(), args);
    console.log(chalk_1.default.green(`\n${args.database} database has been created.`));
    console.log(chalk_1.default.cyanBright(`=> ${root("src", "app", "database.ts")}`));
    console.timeEnd("duration");
})
    .command("namespace <name>", "make a namespace", (yargs) => {
    yargs.positional("name", {
        describe: "namespace name",
        type: "string",
    });
}, async (args) => {
    console.time("duration");
    if (!(await isValidRoot()))
        return;
    const namespacePath = root("src", "namespaces", args.name + ".ts");
    await promises_1.default.writeFile(namespacePath, `export function getSome${
    // @ts-ignore
    args.name[0].toUpperCase() + args.name.slice(1)}Value(): unknown {}`, "utf8");
    const appFile = await promises_1.default.readFile(root("src", "app.ts"), "utf8");
    const appLines = appFile.split("\n");
    if (/^\s*$/.test(appLines[appLines.length - 1]))
        appLines.pop();
    const spaceIndex = appLines.findIndex((line) => /^\s*$/.test(line));
    appLines.splice(spaceIndex, 0, `export * from "./namespaces/${args.name}.js"`);
    appLines.push(`export * as ${args.name} from "./namespaces/${args.name}.js"`);
    await promises_1.default.writeFile(root("src", "app.ts"), appLines.join("\n"), "utf8");
    console.log(chalk_1.default.green(`\n${args.name} namespace has been created.`));
    console.log(chalk_1.default.cyanBright(`=> ${namespacePath}`));
    console.timeEnd("duration");
})
    .command("table <name>", "make a database table", (yargs) => {
    yargs.positional("name", {
        describe: "table name",
        type: "string",
    });
}, async (args) => {
    console.time("duration");
    if (!(await isValidRoot()))
        return;
    const tablePath = root("src", "tables", args.name + ".ts");
    const template = await promises_1.default.readFile((0, path_1.join)(__dirname, "..", "templates", "table"), "utf8");
    await promises_1.default.writeFile(tablePath, template
        // @ts-ignore
        .replace(/{{ name }}/g, args.name)
        .replace(/{{ Name }}/g, 
    // @ts-ignore
    args.name[0].toUpperCase() + args.name.slice(1)), "utf8");
    console.log(chalk_1.default.green(`\n${args.name} table has been created.`));
    console.log(chalk_1.default.cyanBright(`=> ${tablePath}`));
    console.timeEnd("duration");
})
    .help().argv;
function makeFile(id, arg) {
    return [
        `${id} <${arg}>`,
        "create bot " + id,
        (yargs) => {
            yargs.positional(arg, {
                describe: id + " " + arg,
            });
            if (id === "listener") {
                yargs.option("name", {
                    alias: "n",
                    describe: "The name of listener file",
                    default: null,
                });
            }
        },
        async (argv) => {
            console.time("duration");
            const filename = (argv.name ?? argv[arg]) + ".ts";
            if (arg === "event") {
                if (!argv[arg]) {
                    return console.error(chalk_1.default.red("you should give a Discord Client event name."));
                }
                const event = Object.keys(events).find((key) => key.toLowerCase() === argv[arg].toLowerCase());
                if (event) {
                    argv[arg] = event;
                }
                else {
                    console.error(chalk_1.default.red("you should give a valid Discord Client event name."));
                    for (const event of Object.keys(events)) {
                        const similarity = string_similarity_1.default.compareTwoStrings(event.toLowerCase(), argv[arg].toLowerCase());
                        if (similarity > 0.75) {
                            console.log(`Did you mean ${chalk_1.default.cyanBright(event)} instead?`);
                        }
                    }
                    return;
                }
            }
            if (!(await isValidRoot()))
                return;
            const { style } = require(root("package.json"));
            const template = await promises_1.default.readFile((0, path_1.join)(__dirname, "..", "templates", id + (id === "command" && style === "chain" ? "_chain" : "")), "utf8");
            let file = template.replace(new RegExp(`{{ ${arg} }}`, "g"), argv[arg]);
            if (arg === "event") {
                let args = events[argv[arg]];
                args = typeof args === "string" ? args : args.join(", ");
                file = file.replace(`{{ args }}`, args);
            }
            const directory = root("src", id + "s");
            if (!fs_1.default.existsSync(directory)) {
                console.warn(`${id}s directory not exists.`);
                await promises_1.default.mkdir(directory, { recursive: true });
                console.log(chalk_1.default.green(`${id}s directory created.`));
                console.log(chalk_1.default.cyanBright(`=> ${directory}`));
            }
            const path = (0, path_1.join)(directory, filename);
            if (fs_1.default.existsSync(path))
                return console.error(chalk_1.default.red(`${argv[arg]} ${id} already exists.`));
            await promises_1.default.writeFile(path, file, "utf8");
            console.log(chalk_1.default.green(`\n${argv[arg]} ${id} has been created.`));
            console.log(chalk_1.default.cyanBright(`=> ${path}`));
            console.timeEnd("duration");
        },
    ];
}
